/* DISK SCHEDULER SCAN */
#include <stdio.h>
#include <stdlib.h>
#include <signal.h>
#include <unistd.h>
#include <errno.h>

#define TRUE 1
#define FALSE 0
#define NPROC 10
#define MAX_TRACCE 100

typedef int *Tsync_buf;

typedef struct Tlista {
  int key;
  int pid;
  int traccia;
  int num_proc;
  struct Tlista *next;
} Tlista;

Tlista *ph_up, *ph_down, *pcur, *prec, *pnew;

Tsync_buf sync_buf[NPROC], coda;
int num_proc;

int childpid, ordine, pid, traccia;
int up, headpos, busy;
int proc_pid, proc_num_proc, proc_traccia;

Tlista *
inserisci (Tlista * ph, int key)
{
  pcur = ph;
  prec = pcur;
  while (pcur->key < key) {	/* cerca il suo posto  */
    prec = pcur;
    pcur = pcur->next;
  }
  pnew = (Tlista *) malloc (sizeof (Tlista));	/* inserimento ordinato */
  pnew->next = prec->next;
  prec->next = pnew;
  pnew->key = key;
  return pnew;
}

Tsync_buf
make_sync_buf ()
{
  int *sync_buf;

  sync_buf = calloc (2, sizeof (int));
  pipe (sync_buf);
  return sync_buf;
}

void
estrae ()
{
  if (ph_up->next->key == 10000 && ph_down->next->key == 10000)	/* code vuote */
    busy = FALSE;
  else {
    /*  controlla se deve cambiare direzione perche' la coda attuale e' vuota */
    if (up)
      if (ph_up->next->key < 10000)
	prec = ph_up;
      else {
	up = FALSE;
	prec = ph_down;
      }
    else
      /* down */ if (ph_down->next->key < 10000)
      prec = ph_down;
    else {
      up = TRUE;
      prec = ph_up;
    }

/* estrae sempre dalla testa di una coda che contiene almeno un elemento */
    proc_pid = prec->next->pid;
    proc_num_proc = prec->next->num_proc;
    proc_traccia = prec->next->traccia;
    pcur = prec->next->next;
    free (prec->next);
    prec->next = pcur;
    headpos = proc_traccia;
    ordine++;
    fprintf (stderr, "\n\t\t\tDirezione %s\n ", up ? "UP" : "DOWN");

    alarm (4);
  }
}

void
scheduler_alarm ()
{
  if (write (sync_buf[proc_num_proc][1], &proc_traccia, sizeof (int)) <= 0) {
    fprintf (stderr, "ERRORE : scrittura su pipe %d\n", num_proc);
    exit (1);
  }
  if (write (sync_buf[proc_num_proc][1], &ordine, sizeof (int)) <= 0) {
    fprintf (stderr, "ERRORE : scrittura su pipe %d\n", num_proc);
    exit (1);
  }
  printf ("invio segnale a pid: %d num_proc: %d traccia: %d ordine:%d\n",
	  proc_pid, proc_num_proc, proc_traccia, ordine);
  kill (proc_pid, SIGUSR1);

  estrae ();
  signal (SIGALRM, scheduler_alarm);
}

void
user_alarm ()
{
  signal (SIGALRM, user_alarm);
}

void
user_usr1 ()
{
  int traccia, ordine;

  if (read (sync_buf[num_proc][0], &traccia, sizeof (int)) <= 0) {
    fprintf (stderr, "ERRORE : signal\n");
    exit (1);
  }
  if (read (sync_buf[num_proc][0], &ordine, sizeof (int)) <= 0) {
    fprintf (stderr, "ERRORE : signal\n");
    exit (1);
  }
  fprintf (stderr, "Processo: %d\tOrdine: %d\tTraccia: %d\n",
	   getpid (), ordine, traccia);
  signal (SIGUSR1, user_usr1);
}

void
processo_utente ()
{
  int i, somma, pid, attesa, traccia;

  setbuf (stdout, 0);
  signal (SIGUSR1, user_usr1);
  signal (SIGALRM, user_alarm);
  pid = getpid ();
  fprintf (stderr, "Nasce il processo %d num_proc: %d\n", pid, num_proc);
  srand (pid);

  while (1) {
    somma = 0;
    for (i = 0; i < 10000; i++)
      somma += attesa;		/* fa altro */
    attesa = 1 + (int) (10.0 * rand () / (RAND_MAX + 1.0));
    alarm (attesa);
    pause ();			/*  attesa tempo casuale */
    traccia = 1 + (int) ((float) MAX_TRACCE * rand () / (RAND_MAX + 1.0));
    printf ("Processo: %d Richiesta traccia: %d dopo aver atteso %d secondi\n",
	    getpid (), traccia, attesa);
    if (write (coda[1], &pid, sizeof (int)) <= 0) {
      fprintf (stderr, "ERRORE : scrittura su coda %d\n", num_proc);
      exit (1);
    }
    if (write (coda[1], &num_proc, sizeof (int)) <= 0) {
      fprintf (stderr, "ERRORE : scrittura su coda %d\n", num_proc);
      exit (1);
    }
    if (write (coda[1], &traccia, sizeof (int)) <= 0) {
      fprintf (stderr, "ERRORE : scrittura su coda %d\n", num_proc);
      exit (1);
    }
    pause ();			/*  attesa servizio */
  }
}


int
main ()
{

  int childpid, n;
  sigset_t newset, *newset_p;

  setbuf (stdout, 0);

  pnew = (Tlista *) malloc (sizeof (Tlista));
  pnew->next = NULL;
  pnew->key = 10000;		/* sentinella finale comune alle due code */

  ph_up = (Tlista *) malloc (sizeof (Tlista));
  ph_up->next = pnew;
  ph_up->key = -10000;		/* sentinella iniziale */

  ph_down = (Tlista *) malloc (sizeof (Tlista));
  ph_down->next = pnew;
  ph_down->key = -10000;	/* sentinella iniziale */

  up = TRUE;
  headpos = 0;
  busy = FALSE;
  coda = make_sync_buf ();
  for (num_proc = 0; num_proc < NPROC; num_proc++)
    sync_buf[num_proc] = make_sync_buf ();

  for (num_proc = 0; num_proc < NPROC; num_proc++)
    if ((childpid = fork ()) == 0)
      processo_utente ();

  /*      padre - scheduler  */

  newset_p = &newset;
  sigemptyset (newset_p);
  sigaddset (newset_p, SIGALRM);

  signal (SIGALRM, scheduler_alarm);
  while (1) {
    while ((n=read (coda[0], &pid, sizeof (int))) < 0 && errno == EINTR);
    if (n <=0 && errno != EINTR) {  
      fprintf (stderr, "scheduler ERRORE : lettura da pipe %d \n", errno);
      exit (1);
    }
    while ((n=read (coda[0], &num_proc, sizeof (int))) < 0 && errno == EINTR);
    if (n <=0 && errno != EINTR) {  
      fprintf (stderr, "scheduler ERRORE : lettura da pipe %d \n", errno);
      exit (1);
    }
    while ((n=read (coda[0], &traccia, sizeof (int))) < 0 && errno == EINTR);
    if (n <=0 && errno != EINTR) {  
      fprintf (stderr, "scheduler ERRORE : lettura da pipe %d \n", errno);
      exit (1);
    }

    /* accoda */
    sigprocmask (SIG_BLOCK, newset_p, NULL);
    if (up)
      if (traccia > headpos)
	pcur = inserisci (ph_up, traccia);
      else
	pcur = inserisci (ph_down, MAX_TRACCE - traccia);
    else
      /* movimento down */
    if (traccia > headpos)
      pcur = inserisci (ph_up, traccia);
    else
      pcur = inserisci (ph_down, MAX_TRACCE - traccia);
    pcur->traccia = traccia;
    pcur->pid = pid;
    pcur->num_proc = num_proc;

    pcur = ph_up->next;
    fprintf (stderr, "\nCoda UP, HEADPPOS : %d\n ", headpos);
    while (pcur->next != NULL) {
      fprintf (stderr, "key %d pid %d num_proc %d traccia %d\n",
	       pcur->key, pcur->pid, pcur->num_proc, pcur->traccia);
      pcur = pcur->next;
    }
    pcur = ph_down->next;
    fprintf (stderr, "\nCoda DOWN :\n ");
    while (pcur->next != NULL) {
      fprintf (stderr, "key %d pid %d num_proc %d traccia %d\n",
	       pcur->key, pcur->pid, pcur->num_proc, pcur->traccia);
      pcur = pcur->next;
    }

    if (!busy) {
      busy = TRUE;
      estrae ();		/* implica attesa timeout */
    }
    sigprocmask (SIG_UNBLOCK, newset_p, NULL);
  }
}
