/******************************************************************************
 *                                                                            *
 *  Questo file raccoglie due funzioni richiamabili da un programma FORTRAN   *
 *  che permettono a piu' processi di leggere lo stesso file in mutua         *
 *  esclusione. Il file deve contenere una lista di numeri, ciascuno seguito  *
 *  da una parola chiave definita come GET_IT qui sotto, che viene prima      *
 *  cambiata in WORKING da una chiamata a 'next(file_lista)' (che ritorna il  *
 *  prossimo numero da processare), e poi cambiata in DONE da una chiamata a  *
 *  'made(file_lista,numero)'. Tutto cio' permette di continuare un lavoro    *
 *  eventualmente interrotto in questo modo:                                  *
 *  1) editare il file lista e reinpostare tutti i WORKING come GET_IT;       *
 *  2) cancellare gli eventuali file  'LCK?????' presenti e soprattutto       *
 *     il file nome_lista.lock !!!                                            *
 *                                                                            *
 *  N.B.: Le funzioni devono essere definite in Fortran come REAL*4.          *
 *                                                                            *
 ******************************************************************************/

#include <stdio.h>
#include <string.h>
#include <errno.h>
#include <fcntl.h>

/* Warning: GET_IT,WORKING & DONE's lenght must be exactly LEN ! */
#define LEN 5
#define GET_IT "getit"
#define WORKING "Workg"
#define DONE "DONE!"

#define BUFSIZE 75
#define NCHAR 80

extern int errno;

int    my_lock(lockfile)
char *lockfile;
{
   int     tempfd;
   char    tempfile[NCHAR], buff[NCHAR];

   strcpy(buff,lockfile);
   strcat(buff,".lock");

   sprintf(tempfile, "LCK%d", getpid());
   if ((tempfd = creat(tempfile, 0444)) < 0) {
      printf("LOCK ERROR: Can't creat temp file");
      return -1;
   }
   close(tempfd);

   while (symlink(tempfile, buff) < 0) {
      if (errno != EEXIST) {
	 printf("LOCK ERROR: Link error");
         return -1;
      }
      sleep(1);
   }
   if (unlink(tempfile) < 0) {
      printf("LOCK ERROR: Unlink error for tempfile");
      return -1;
   }
   return 0;
}

int    my_unlock(lockfile)
char *lockfile;
{
   char buff[NCHAR];

   strcpy(buff,lockfile);
   strcat(buff,".lock");

   if (unlink(buff) < 0) {
      printf("Unlink error for LOCKFILE '%s'",lockfile);
      return -1;
   }
   return 0;
}


/* NPID: torna il program id */

long int    npid_()
{
   return getpid();
}


/*
 * NEXT(nomefile): sostituisce il primo GET_IT di 'nomefile' con WORKING. 
 * 
 * Valore di ritorno: 0 = fine file, n = prossimo numero, -1 = errore
 */

long int     next_(file_di_lock)
char *file_di_lock;
{
   char   *s, buf[BUFSIZE], file_name[NCHAR];
   int     n;
   long    error = -1L, zero = 0L;
   long    numero, mem;
   FILE   *fp;

   n= -1; /* convert FORTRAN string */
   while(!isspace(file_di_lock[++n]))
      file_name[n] = file_di_lock[n];
   file_name[n] = '\0';

   if (my_lock(file_name) != 0)
      return error;

   if ((fp = fopen(file_name, "r+")) == NULL)
      return error;
   do {
      if ((mem = ftell(fp)) == -1L)
	 return error;
      if (fgets(buf, BUFSIZE, fp) == NULL) {
         if (my_unlock(file_name) != 0)
            return error;
	 return zero;
      }
   } while ((s = strstr(buf, GET_IT)) == NULL);
   sscanf(buf, "%i ", &numero);
   strncpy(s, WORKING, LEN);
   if (fseek(fp, mem, 0) != 0)
      return error;
   if (fputs(buf, fp) == EOF)
      return error;
   if (fclose(fp) == EOF)
      return error;

   if (my_unlock(file_name) != 0)
      return error;

   return numero;
}


/*
 * DONE(nomefile,n): sostituisce DONE a WORKING in 'nomefile' accanto a 'n'. 
 * 
 * Valore di ritorno: 0 = tutto ok , -1 = errore
 */

long int     done_(file_di_lock, number)
char *file_di_lock;
long int *number;
{
   char    buf[BUFSIZE], file_name[NCHAR];
   int     n;
   long    numero, error = -1L, zero = 0L;
   long    mem;
   FILE   *fp;

   n= -1; /* convert FORTRAN string */
   while(!isspace(file_di_lock[++n]))
      file_name[n] = file_di_lock[n];
   file_name[n] = '\0';

   if (my_lock(file_name) != 0)
      return error;

   if ((fp = fopen(file_name, "r+")) == NULL)
      return error;
   do {
      if ((mem = ftell(fp)) == -1L)
	 return error;
      if (fgets(buf, BUFSIZE, fp) == NULL)
	 return error;
      sscanf(buf, "%i ", &numero);
   } while (numero != (*number));
   strncpy(strstr(buf, WORKING), DONE, LEN);
   if (fseek(fp, mem, 0) != 0)
      return error;
   if (fputs(buf, fp) == EOF)
      return error;
   if (fclose(fp) == EOF)
      return error;

   if (my_unlock(file_name) != 0)
      return error;

   return zero;
}
