/* read "n" bytes from a stream socket */

int 
readn(fd, ptr, nbytes)
  register int    fd;
  register char  *ptr;
  register int    nbytes;
{
  int             nleft, nread;

  nleft = nbytes;
  while (nleft > 0) {
    nread = read(fd, ptr, nleft);
    if (nread < 0)
      return (nread);		/* error */
    else if (nread == 0)
      break;			/* EOF */
    nleft -= nread;
    ptr += nread;
  }
  return (nbytes - nleft);
}


/* write "n" bytes to a stream socket */

int 
writen(fd, ptr, nbytes)
  register int    fd;
  register char  *ptr;
  register int    nbytes;
{
  int             nleft, nwritten;

  nleft = nbytes;
  while (nleft > 0) {
    nwritten = write(fd, ptr, nleft);
    if (nwritten < 0)
      return (nwritten);	/* error */
    else if (nwritten == 0)
      break;			/* EOF */
    nleft -= nwritten;
    ptr += nwritten;
  }
  return (nbytes - nleft);
}


/* read a line from a stream socket */

int 
readline(fd, ptr, maxlen)
  register int    fd;
  register char  *ptr;
  register int    maxlen;
{
  int             n, rc;
  char            c;

  for (n = 1; n < maxlen; n++) {
    if ((rc = read(fd, &c, 1)) == 1) {
      *ptr++ = c;
      if (c == '\n')
	break;
    } else if (rc == 0) {
      if (n == 1)
	return (0);		/* EOF */
      else
	break;
    } else
      return (-1);		/* error */
  }
  *ptr = 0;
  return (n);
}

#define MAXLINE 512

str_echo(sockfd)
  int             sockfd;
{
  int             i, n;
  char            line[MAXLINE];

  for (;;) {
    n = readline(sockfd, line, MAXLINE);
    if (n == 0)
      return;			/* connection terminated */
    else if (n < 0)
      err_dump("str_echo: readline error");
    for (i = 0; i <=n; i++) 
      line[i] = toupper(line[i]);
    if (writen(sockfd, line, n) != n)
      err_dump("str_echo: writen error");
  }
}

#include <stdio.h>
#define MAXLINE 512

str_client(fp, sockfd)
  register FILE  *fp;
  register int    sockfd;
{

  int             n;
  char            sendline[MAXLINE], recvline[MAXLINE + 1];

  while (fgets(sendline, MAXLINE, fp) != NULL) {
    n = strlen(sendline);
    if (writen(sockfd, sendline, n) != n)
      err_sys("str_client: writen error on socket");

    n = readline(sockfd, recvline, MAXLINE);
    if (n < 0)
      err_dump("str_client: readline error");
    recvline[n] = 0;
    fputs(recvline, stdout);
  }
  if (ferror(fp))
    err_sys("str_client: error reading file");
}

#include <sys/types.h>
#include <sys/socket.h>

#define MAXMSG 2048

dg_echo(sockfd, pcli_addr, maxclilen)
  int             sockfd;
  struct sockaddr *pcli_addr;
  int maxclilen;
{
  int             i, n, clilen;
  char            mesg[MAXMSG];

  for (;;) {
    clilen = maxclilen;
    n = recvfrom(sockfd, mesg, MAXMSG, 0, pcli_addr, &clilen);
    if (n < 0)
      err_dump("dg_echo: recvfrom error");
    for (i = 0; i <=n; i++) 
      mesg[i] = toupper(mesg[i]);
    if (sendto(sockfd, mesg, n, 0, pcli_addr, clilen) != n)
      err_dump("dg_echo: sendto error");
  }
}

#define MAXLINE 512

dg_client(fp, sockfd, pserv_addr, servlen)
  FILE  *fp;
  int    sockfd;
  struct sockaddr *pserv_addr;
  int servlen;
{

  int             n;
  char            sendline[MAXLINE], recvline[MAXLINE + 1];

  while (fgets(sendline, MAXLINE, fp) != NULL) {
    n = strlen(sendline);
    if (sendto(sockfd, sendline, n, 0, pserv_addr, servlen) != n)
      err_dump("dg_client: sendto error on socket");

    n = recvfrom(sockfd, recvline, MAXLINE, 0,
                         (struct sockaddr *) 0, (int *) 0);
    if (n < 0)
      err_dump("dg_client: readline error");
    recvline[n] = 0;
    fputs(recvline, stdout);
  }
  if (ferror(fp))
    err_sys("dg_client: error reading file");
}
